import enum


class TaskType(enum.Enum):
    classification = 0
    regression = 1

    @staticmethod
    def from_str(task_type: str):
        if task_type == "classification":
            return TaskType.classification
        elif task_type == "regression":
            return TaskType.regression
        else:
            raise ValueError("Invalid task type: {}".format(task_type))

    @staticmethod
    def list_str():
        return ["classification", "regression"]


class ProblemType(enum.IntEnum):
    binary_classification = 1
    multi_class_classification = 2
    multi_label_classification = 3
    single_column_regression = 4
    multi_column_regression = 5

    @staticmethod
    def from_str(label):
        if label == "binary_classification":
            return ProblemType.binary_classification
        elif label == "multi_class_classification":
            return ProblemType.multi_class_classification
        elif label == "multi_label_classification":
            return ProblemType.multi_label_classification
        elif label == "single_column_regression":
            return ProblemType.single_column_regression
        elif label == "multi_column_regression":
            return ProblemType.multi_column_regression
        else:
            raise NotImplementedError



import os

from fastapi import FastAPI

from .predict import AutoVFPredict


app = FastAPI()
axgp = AutoVFPredict(model_path=os.environ.get("AUTOVF_MODEL_PATH"))
schema = axgp.get_prediction_schema()


@app.post("/predict")
def predict(sample: schema):
    sample = sample.json()
    return axgp.predict_single(sample)



from typing import List, Optional

from pydantic import BaseModel

from .enums import ProblemType


class ModelConfig(BaseModel):
    train_filename: str
    test_filename: Optional[str] = None
    idx: str
    targets: List[str]
    problem_type: ProblemType
    output: str
    features: List[str]
    num_folds: int
    use_gpu: bool
    seed: int
    categorical_features: List[str]
    num_trials: int
    time_limit: Optional[int] = None
    fast: bool
    project_name: Optional[str] = None
    project_location: Optional[str] = None



import os
from dataclasses import dataclass
from typing import List, Optional

import joblib
import numpy as np
import pandas as pd
from sklearn.model_selection import KFold, StratifiedKFold
from sklearn.preprocessing import LabelEncoder, OrdinalEncoder
from sklearn.utils.multiclass import type_of_target

from .enums import ProblemType
from .logger import logger
from .schemas import ModelConfig
from .utils import predict_model, reduce_memory_usage, train_model


@dataclass
class AutoVF:
    # required arguments
    train_filename: str
    output: str

    # optional arguments
    test_filename: Optional[str] = None
    task: Optional[str] = None
    idx: Optional[str] = "id"
    targets: Optional[List[str]] = None
    features: Optional[List[str]] = None
    categorical_features: Optional[List[str]] = None
    use_gpu: Optional[bool] = False
    num_folds: Optional[int] = 5
    seed: Optional[int] = 42
    num_trials: Optional[int] = 1000
    time_limit: Optional[int] = None
    fast: Optional[bool] = False
    project_name: Optional[str] = None
    project_location: Optional[str] = None

    def __post_init__(self):
        if os.path.exists(self.output):
            raise Exception(
                "Output directory already exists. Please specify some other directory."
            )
        os.makedirs(self.output, exist_ok=True)
        logger.info(f"Output directory: {self.output}")

        if self.targets is None:
            logger.warning("No target columns specified. Will default to `target`.")
            self.targets = ["target"]

        if self.idx is None:
            logger.warning("No id column specified. Will default to `id`.")
            self.idx = "id"

    def _create_folds(self, train_df, problem_type):
        if "kfold" in train_df.columns:
            self.num_folds = len(np.unique(train_df["kfold"]))
            logger.info("Using `kfold` for folds from training data")
            return train_df

        logger.info("Creating folds")
        train_df["kfold"] = -1
        if problem_type in (
            ProblemType.binary_classification,
            ProblemType.multi_class_classification,
        ):
            y = train_df[self.targets].values
            kf = StratifiedKFold(
                n_splits=self.num_folds, shuffle=True, random_state=self.seed
            )
            for fold, (_, valid_indicies) in enumerate(kf.split(X=train_df, y=y)):
                train_df.loc[valid_indicies, "kfold"] = fold
        elif problem_type == ProblemType.single_column_regression:
            y = train_df[self.targets].values
            num_bins = int(np.floor(1 + np.log2(len(train_df))))
            if num_bins > 10:
                num_bins = 10
            kf = StratifiedKFold(
                n_splits=self.num_folds, shuffle=True, random_state=self.seed
            )
            train_df["bins"] = pd.cut(
                train_df[self.targets].values.ravel(), bins=num_bins, labels=False
            )
            for fold, (_, valid_indicies) in enumerate(
                kf.split(X=train_df, y=train_df.bins.values)
            ):
                train_df.loc[valid_indicies, "kfold"] = fold
            train_df = train_df.drop("bins", axis=1)
        elif problem_type == ProblemType.multi_column_regression:
            y = train_df[self.targets].values
            kf = KFold(n_splits=self.num_folds, shuffle=True, random_state=self.seed)
            for fold, (_, valid_indicies) in enumerate(kf.split(X=train_df, y=y)):
                train_df.loc[valid_indicies, "kfold"] = fold
        # TODO: use iterstrat
        elif problem_type == ProblemType.multi_label_classification:
            y = train_df[self.targets].values
            kf = KFold(n_splits=self.num_folds, shuffle=True, random_state=self.seed)
            for fold, (_, valid_indicies) in enumerate(kf.split(X=train_df, y=y)):
                train_df.loc[valid_indicies, "kfold"] = fold
        else:
            raise Exception("Problem type not supported")
        return train_df

    def _determine_problem_type(self, train_df):
        if self.task is not None:
            if self.task == "classification":
                if len(self.targets) == 1:
                    if len(np.unique(train_df[self.targets].values)) == 2:
                        problem_type = ProblemType.binary_classification
                    else:
                        problem_type = ProblemType.multi_class_classification
                else:
                    problem_type = ProblemType.multi_label_classification

            elif self.task == "regression":
                if len(self.targets) == 1:
                    problem_type = ProblemType.single_column_regression
                else:
                    problem_type = ProblemType.multi_column_regression
            else:
                raise Exception("Problem type not understood")

        else:
            target_type = type_of_target(train_df[self.targets].values)
            # target type is one of the following using scikit-learn's type_of_target
            # * 'continuous': `y` is an array-like of floats that are not all
            #   integers, and is 1d or a column vector.
            # * 'continuous-multioutput': `y` is a 2d array of floats that are
            #   not all integers, and both dimensions are of size > 1.
            # * 'binary': `y` contains <= 2 discrete values and is 1d or a column
            #   vector.
            # * 'multiclass': `y` contains more than two discrete values, is not a
            #   sequence of sequences, and is 1d or a column vector.
            # * 'multiclass-multioutput': `y` is a 2d array that contains more
            #   than two discrete values, is not a sequence of sequences, and both
            #   dimensions are of size > 1.
            # * 'multilabel-indicator': `y` is a label indicator matrix, an array
            #   of two dimensions with at least two columns, and at most 2 unique
            #   values.
            # * 'unknown': `y` is array-like but none of the above, such as a 3d
            #   array, sequence of sequences, or an array of non-sequence objects.
            if target_type == "continuous":
                problem_type = ProblemType.single_column_regression
            elif target_type == "continuous-multioutput":
                problem_type = ProblemType.multi_column_regression
            elif target_type == "binary":
                problem_type = ProblemType.binary_classification
            elif target_type == "multiclass":
                problem_type = ProblemType.multi_class_classification
            elif target_type == "multilabel-indicator":
                problem_type = ProblemType.multi_label_classification
            else:
                raise Exception(
                    "Unable to infer `problem_type`. Please provide `classification` or `regression`"
                )
        logger.info(f"Problem type: {problem_type.name}")
        return problem_type

    def _inject_idxumn(self, df):
        if self.idx not in df.columns:
            df[self.idx] = np.arange(len(df))
        return df

    def _process_data(self):
        logger.info("Reading training data")
        train_df = pd.read_csv(self.train_filename)
        train_df = reduce_memory_usage(train_df)
        problem_type = self._determine_problem_type(train_df)

        train_df = self._inject_idxumn(train_df)
        if self.test_filename is not None:
            test_df = pd.read_csv(self.test_filename)
            test_df = reduce_memory_usage(test_df)
            test_df = self._inject_idxumn(test_df)

        # create folds
        train_df = self._create_folds(train_df, problem_type)
        ignore_columns = [self.idx, "kfold"] + self.targets

        if self.features is None:
            self.features = list(train_df.columns)
            self.features = [x for x in self.features if x not in ignore_columns]

        # encode target(s)
        if problem_type in [
            ProblemType.binary_classification,
            ProblemType.multi_class_classification,
        ]:
            logger.info("Encoding target(s)")
            target_encoder = LabelEncoder()
            target_encoder.fit(
                train_df[self.targets].values.reshape(
                    -1,
                )
            )
            train_df.loc[:, self.targets] = target_encoder.transform(
                train_df[self.targets].values.reshape(
                    -1,
                )
            )
        else:
            target_encoder = None

        if self.categorical_features is None:
            # find categorical features
            categorical_features = []
            for col in self.features:
                if train_df[col].dtype == "object":
                    categorical_features.append(col)

        else:
            categorical_features = self.categorical_features

        logger.info(f"Found {len(categorical_features)} categorical features.")

        if len(categorical_features) > 0:
            logger.info("Encoding categorical features")
        categorical_encoders = {}
        for fold in range(self.num_folds):
            fold_train = train_df[train_df.kfold != fold].reset_index(drop=True)
            fold_valid = train_df[train_df.kfold == fold].reset_index(drop=True)
            if self.test_filename is not None:
                test_fold = test_df.copy(deep=True)
            if len(categorical_features) > 0:
                ord_encoder = OrdinalEncoder(
                    handle_unknown="use_encoded_value", unknown_value=np.nan
                )
                fold_train[categorical_features] = ord_encoder.fit_transform(
                    fold_train[categorical_features].values
                )
                fold_valid[categorical_features] = ord_encoder.transform(
                    fold_valid[categorical_features].values
                )
                if self.test_filename is not None:
                    test_fold[categorical_features] = ord_encoder.transform(
                        test_fold[categorical_features].values
                    )
                categorical_encoders[fold] = ord_encoder
            fold_train.to_feather(
                os.path.join(self.output, f"train_fold_{fold}.feather")
            )
            fold_valid.to_feather(
                os.path.join(self.output, f"valid_fold_{fold}.feather")
            )
            if self.test_filename is not None:
                test_fold.to_feather(
                    os.path.join(self.output, f"test_fold_{fold}.feather")
                )

        # save config
        model_config = {}
        model_config["idx"] = self.idx
        model_config["features"] = self.features
        model_config["categorical_features"] = categorical_features
        model_config["train_filename"] = self.train_filename
        model_config["test_filename"] = self.test_filename
        model_config["output"] = self.output
        model_config["problem_type"] = problem_type
        model_config["idx"] = self.idx
        model_config["targets"] = self.targets
        model_config["use_gpu"] = self.use_gpu
        model_config["num_folds"] = self.num_folds
        model_config["seed"] = self.seed
        model_config["num_trials"] = self.num_trials
        model_config["time_limit"] = self.time_limit
        model_config["fast"] = self.fast
        model_config["project_name"] = self.project_name
        model_config["project_location"] = self.project_location

        self.model_config = ModelConfig(**model_config)
        logger.info(f"Model config: {self.model_config}")
        logger.info("Saving model config")
        joblib.dump(self.model_config, f"{self.output}/avf.config")

        # save encoders
        logger.info("Saving encoders")
        joblib.dump(categorical_encoders, f"{self.output}/avf.categorical_encoders")
        joblib.dump(target_encoder, f"{self.output}/avf.target_encoder")

    def train(self):
        self._process_data()
        best_params = train_model(self.model_config)
        logger.info("Training complete")
        self.predict(best_params)

    def predict(self, best_params):
        logger.info("Creating OOF and test predictions")
        predict_model(self.model_config, best_params)



import json
import os
from dataclasses import dataclass
from typing import Dict, Union

import joblib
import numpy as np
import pandas as pd
import xgboost as xgb
from pydantic import create_model

from .enums import ProblemType
from .utils import fetch_xgb_model_params, reduce_memory_usage


xgb.set_config(verbosity=0)


@dataclass
class AutoVFPredict:
    model_path: str

    def __post_init__(self):
        self.model_config = joblib.load(os.path.join(self.model_path, "avf.config"))
        self.target_encoder = joblib.load(
            os.path.join(self.model_path, "avf.target_encoder")
        )
        self.categorical_encoders = joblib.load(
            os.path.join(self.model_path, "avf.categorical_encoders")
        )
        self.models = []
        for fold in range(self.model_config.num_folds):
            model_ = joblib.load(os.path.join(self.model_path, f"avf_model.{fold}"))
            self.models.append(model_)

        _, self.use_predict_proba, _, _ = fetch_xgb_model_params(self.model_config)

    # TODO: improve schema with optional null values
    def get_prediction_schema(self):
        cat_features = self.model_config.categorical_features
        schema = {"PredictSchema": {}}
        for cf in cat_features:
            schema["PredictSchema"][cf] = "str"

        for feat in self.model_config.features:
            if feat not in cat_features:
                schema["PredictSchema"][feat] = 10.0
        return create_model("PredictSchema", **schema["PredictSchema"])

    def _predict_df(self, df):
        categorical_features = self.model_config.categorical_features
        final_preds = []
        test_ids = df[self.model_config.idx].values
        for fold in range(self.model_config.num_folds):
            fold_test = df.copy(deep=True)
            if len(categorical_features) > 0:
                categorical_encoder = self.categorical_encoders[fold]
                fold_test[categorical_features] = categorical_encoder.transform(
                    fold_test[categorical_features].values
                )

            test_features = fold_test[self.model_config.features]

            for col in test_features.columns:
                if test_features[col].dtype == "object":
                    test_features[col] = test_features[col].astype(np.int64)

            if self.model_config.problem_type in (
                ProblemType.multi_column_regression,
                ProblemType.multi_label_classification,
            ):
                test_preds_mll = []
                for midx in range(len(self.models[fold])):
                    if (
                        self.model_config.problem_type
                        == ProblemType.multi_column_regression
                    ):
                        test_pred_temp = self.models[fold][midx].predict(test_features)
                    else:
                        test_pred_temp = self.models[fold][midx].predict_proba(
                            test_features
                        )[:, 1]
                    test_preds_mll.append(test_pred_temp)

                test_preds = np.column_stack(test_preds_mll)
            else:
                if self.use_predict_proba:
                    test_preds = self.models[fold].predict_proba(test_features)
                else:
                    test_preds = self.models[fold].predict(test_features)

            final_preds.append(test_preds)

        final_preds = np.mean(final_preds, axis=0)
        if self.target_encoder is None:
            final_preds = pd.DataFrame(
                final_preds, columns=self.model_config.target_cols
            )
        else:
            final_preds = pd.DataFrame(
                final_preds, columns=list(self.target_encoder.classes_)
            )
        final_preds.insert(loc=0, column=self.model_config.idx, value=test_ids)
        return final_preds

    def predict_single(
        self,
        sample: Dict[str, Union[str, int, float]] = None,
        fast_predict: bool = True,
    ):
        sample = json.loads(sample)
        sample_df = pd.DataFrame.from_dict(sample, orient="index").T
        sample_df[self.model_config.idx] = 0
        preds = self._predict_df(sample_df)
        preds = preds.to_dict(orient="records")[0]
        return preds

    def predict_file(self, test_filename: str, out_filename: str):
        test_df = pd.read_csv(test_filename)
        test_df = reduce_memory_usage(test_df)
        final_preds = self._predict_df(test_df)
        final_preds.to_csv(out_filename, index=False)



import os
from argparse import ArgumentParser

import uvicorn

from . import BaseCommand


def serve_autovf_command_factory(args):
    return ServeAutoVFCommand(
        args.model_path, args.port, args.host, args.workers, args.debug
    )


class ServeAutoVFCommand(BaseCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        _parser = parser.add_parser("serve", help="Serve AutoVF API")
        _parser.add_argument(
            "--model_path", help="Path to model", required=True, type=str
        )
        _parser.add_argument(
            "--port", help="Port to serve on", default=9999, type=int, required=False
        )
        _parser.add_argument(
            "--host",
            help="Host to serve on",
            default="127.0.0.1",
            type=str,
            required=False,
        )
        _parser.add_argument(
            "--workers", help="Number of workers", default=1, type=int, required=False
        )
        _parser.add_argument(
            "--debug", help="Debug mode", action="store_true", required=False
        )
        _parser.set_defaults(func=serve_autovf_command_factory)

    def __init__(self, model_path, port, host, workers, debug):
        self.model_path = model_path
        self.port = port
        self.host = host
        self.workers = workers
        self.debug = debug

    def execute(self):
        os.environ["AUTOVF_MODEL_PATH"] = self.model_path
        # run app using uvicorn
        uvicorn.run(
            "autovf.api:app",
            host=self.host,
            port=self.port,
            debug=self.debug,
            workers=self.workers,
        )



import argparse

from .. import __version__
from .predict import PredictAutoVFCommand
from .serve import ServeAutoVFCommand
from .train import TrainAutoVFCommand


def main():
    parser = argparse.ArgumentParser(
        "AutoVF CLI",
        usage="autovf <command> [<args>]",
        epilog="For more information about a command, run: `autovf <command> --help`",
    )
    parser.add_argument(
        "--version", "-v", help="Display AutoVF version", action="store_true"
    )

    commands_parser = parser.add_subparsers(help="commands")
    TrainAutoVFCommand.register_subcommand(commands_parser)
    PredictAutoVFCommand.register_subcommand(commands_parser)
    ServeAutoVFCommand.register_subcommand(commands_parser)

    args = parser.parse_args()

    if args.version:
        print(__version__)
        exit(0)

    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)

    command = args.func(args)
    command.execute()


if __name__ == "__main__":
    main()



import copy
import os
import json
import time    
from functools import partial

import joblib
import numpy as np
import optuna
import pandas as pd
import xgboost as xgb

from .enums import ProblemType
from .logger import logger
from .metrics import Metrics
from .params import get_params
from google.cloud import aiplatform


optuna.logging.set_verbosity(optuna.logging.INFO)


def reduce_memory_usage(df, verbose=True):
    # NOTE: Original author of this function is unknown
    # if you know the *original author*, please let me know.
    numerics = ["int8", "int16", "int32", "int64", "float16", "float32", "float64"]
    start_mem = df.memory_usage().sum() / 1024**2
    for col in df.columns:
        col_type = df[col].dtypes
        if col_type in numerics:
            c_min = df[col].min()
            c_max = df[col].max()
            if str(col_type)[:3] == "int":
                if c_min > np.iinfo(np.int8).min and c_max < np.iinfo(np.int8).max:
                    df[col] = df[col].astype(np.int8)
                elif c_min > np.iinfo(np.int16).min and c_max < np.iinfo(np.int16).max:
                    df[col] = df[col].astype(np.int16)
                elif c_min > np.iinfo(np.int32).min and c_max < np.iinfo(np.int32).max:
                    df[col] = df[col].astype(np.int32)
                elif c_min > np.iinfo(np.int64).min and c_max < np.iinfo(np.int64).max:
                    df[col] = df[col].astype(np.int64)
            else:
                if (
                    c_min > np.finfo(np.float16).min
                    and c_max < np.finfo(np.float16).max
                ):
                    df[col] = df[col].astype(np.float16)
                elif (
                    c_min > np.finfo(np.float32).min
                    and c_max < np.finfo(np.float32).max
                ):
                    df[col] = df[col].astype(np.float32)
                else:
                    df[col] = df[col].astype(np.float64)
    end_mem = df.memory_usage().sum() / 1024**2
    if verbose:
        logger.info(
            "Mem. usage decreased to {:.2f} Mb ({:.1f}% reduction)".format(
                end_mem, 100 * (start_mem - end_mem) / start_mem
            )
        )
    return df


def dict_mean(dict_list):
    mean_dict = {}
    for key in dict_list[0].keys():
        mean_dict[key] = sum(d[key] for d in dict_list) / len(dict_list)
    return mean_dict


def save_valid_predictions(
    final_valid_predictions, model_config, target_encoder, output_file_name
):
    final_valid_predictions = pd.DataFrame.from_dict(
        final_valid_predictions, orient="index"
    ).reset_index()
    if target_encoder is None:
        final_valid_predictions.columns = [model_config.idx] + model_config.targets
    else:
        final_valid_predictions.columns = [model_config.idx] + list(
            target_encoder.classes_
        )

    final_valid_predictions.to_csv(
        os.path.join(model_config.output, output_file_name),
        index=False,
    )


def save_test_predictions(
    final_test_predictions, model_config, target_encoder, test_ids, output_file_name
):
    final_test_predictions = np.mean(final_test_predictions, axis=0)
    if target_encoder is None:
        final_test_predictions = pd.DataFrame(
            final_test_predictions, columns=model_config.targets
        )
    else:
        final_test_predictions = pd.DataFrame(
            final_test_predictions, columns=list(target_encoder.classes_)
        )
    final_test_predictions.insert(loc=0, column=model_config.idx, value=test_ids)
    final_test_predictions.to_csv(
        os.path.join(model_config.output, output_file_name),
        index=False,
    )


def fetch_xgb_model_params(model_config):
    if model_config.problem_type == ProblemType.binary_classification:
        xgb_model = xgb.XGBClassifier
        use_predict_proba = True
        direction = "minimize"
        eval_metric = "logloss"
    elif model_config.problem_type == ProblemType.multi_class_classification:
        xgb_model = xgb.XGBClassifier
        use_predict_proba = True
        direction = "minimize"
        eval_metric = "mlogloss"
    elif model_config.problem_type == ProblemType.multi_label_classification:
        xgb_model = xgb.XGBClassifier
        use_predict_proba = True
        direction = "minimize"
        eval_metric = "logloss"
    elif model_config.problem_type == ProblemType.single_column_regression:
        xgb_model = xgb.XGBRegressor
        use_predict_proba = False
        direction = "minimize"
        eval_metric = "rmse"
    elif model_config.problem_type == ProblemType.multi_column_regression:
        xgb_model = xgb.XGBRegressor
        use_predict_proba = False
        direction = "minimize"
        eval_metric = "rmse"
    else:
        raise NotImplementedError

    return xgb_model, use_predict_proba, eval_metric, direction


def optimize(
    trial,
    xgb_model,
    use_predict_proba,
    eval_metric,
    model_config,
):
    params = get_params(trial, model_config)
    early_stopping_rounds = params["early_stopping_rounds"]
    del params["early_stopping_rounds"]

    metrics = Metrics(model_config.problem_type)

    scores = []

    for fold in range(model_config.num_folds):
        train_feather = pd.read_feather(
            os.path.join(model_config.output, f"train_fold_{fold}.feather")
        )
        valid_feather = pd.read_feather(
            os.path.join(model_config.output, f"valid_fold_{fold}.feather")
        )
        xtrain = train_feather[model_config.features]
        xvalid = valid_feather[model_config.features]

        ytrain = train_feather[model_config.targets].values
        yvalid = valid_feather[model_config.targets].values

        # train model
        model = xgb_model(
            random_state=model_config.seed,
            eval_metric=eval_metric,
            use_label_encoder=False,
            **params,
        )

        if model_config.problem_type in (
            ProblemType.multi_column_regression,
            ProblemType.multi_label_classification,
        ):
            ypred = []
            models = [model] * len(model_config.targets)
            for idx, _m in enumerate(models):
                _m.fit(
                    xtrain,
                    ytrain[:, idx],
                    early_stopping_rounds=early_stopping_rounds,
                    eval_set=[(xvalid, yvalid[:, idx])],
                    verbose=False,
                )
                if model_config.problem_type == ProblemType.multi_column_regression:
                    ypred_temp = _m.predict(xvalid)
                else:
                    ypred_temp = _m.predict_proba(xvalid)[:, 1]
                ypred.append(ypred_temp)
            ypred = np.column_stack(ypred)

        else:
            model.fit(
                xtrain,
                ytrain,
                early_stopping_rounds=early_stopping_rounds,
                eval_set=[(xvalid, yvalid)],
                verbose=False,
            )

            if use_predict_proba:
                ypred = model.predict_proba(xvalid)
            else:
                ypred = model.predict(xvalid)

        # calculate metric
        metric_dict = metrics.calculate(yvalid, ypred)

        if model_config.project_name is not None and model_config.project_location is not None:
            aiplatform.init(
                experiment=model_config.output,
                project=model_config.project_name,
                location=model_config.project_location,
            )
            aiplatform.start_run(run=f"study_{int(time.time())}")
            params = model.get_params()
            param_keys = [
                "booster",
                "colsample_bylevel",
                "colsample_bynode",
                "colsample_bytree",
                "gamma",
                "learning_rate",
                "max_depth",
                "min_child_weight",
                "n_estimators",
                "random_state",
                "reg_alpha",
                "reg_lambda",
                "subsample",
                "tree_method",
            ]
            params = {k: v for k, v in params.items() if k in param_keys}
            aiplatform.log_params(params)
            aiplatform.log_metrics(metric_dict)
            aiplatform.end_run()
        scores.append(metric_dict)
        if model_config.fast is True:
            break

    mean_metrics = dict_mean(scores)
    logger.info(f"Metrics: {mean_metrics}")
    return mean_metrics[eval_metric]


def train_model(model_config):
    xgb_model, use_predict_proba, eval_metric, direction = fetch_xgb_model_params(
        model_config
    )

    optimize_func = partial(
        optimize,
        xgb_model=xgb_model,
        use_predict_proba=use_predict_proba,
        eval_metric=eval_metric,
        model_config=model_config,
    )
    db_path = os.path.join(model_config.output, "params.db")
    study = optuna.create_study(
        direction=direction,
        study_name="autovf",
        storage=f"sqlite:///{db_path}",
        load_if_exists=True,
    )
    study.optimize(
        optimize_func, n_trials=model_config.num_trials, timeout=model_config.time_limit
    )
    return study.best_params


def predict_model(model_config, best_params):

    early_stopping_rounds = best_params["early_stopping_rounds"]
    del best_params["early_stopping_rounds"]

    if model_config.use_gpu is True:
        best_params["tree_method"] = "gpu_hist"
        best_params["gpu_id"] = 0
        best_params["predictor"] = "gpu_predictor"

    xgb_model, use_predict_proba, eval_metric, _ = fetch_xgb_model_params(model_config)

    metrics = Metrics(model_config.problem_type)
    test_metrics = Metrics(model_config.problem_type)
    scores = []
    test_scores = []

    final_test_predictions = []
    final_valid_predictions = {}

    target_encoder = joblib.load(f"{model_config.output}/avf.target_encoder")

    for fold in range(model_config.num_folds):
        logger.info(f"Training and predicting for fold {fold}")
        train_feather = pd.read_feather(
            os.path.join(model_config.output, f"train_fold_{fold}.feather")
        )
        valid_feather = pd.read_feather(
            os.path.join(model_config.output, f"valid_fold_{fold}.feather")
        )

        xtrain = train_feather[model_config.features]
        xvalid = valid_feather[model_config.features]

        valid_ids = valid_feather[model_config.idx].values

        if model_config.test_filename is not None:
            test_feather = pd.read_feather(
                os.path.join(model_config.output, f"test_fold_{fold}.feather")
            )
            xtest = test_feather[model_config.features]
            ytest = test_feather[model_config.targets].values
            test_ids = test_feather[model_config.idx].values

        ytrain = train_feather[model_config.targets].values
        yvalid = valid_feather[model_config.targets].values

        model = xgb_model(
            random_state=model_config.seed,
            eval_metric=eval_metric,
            use_label_encoder=False,
            **best_params,
        )

        if model_config.problem_type in (
            ProblemType.multi_column_regression,
            ProblemType.multi_label_classification,
        ):
            ypred = []
            test_pred = []
            trained_models = []
            for idx in range(len(model_config.targets)):
                _m = copy.deepcopy(model)
                _m.fit(
                    xtrain,
                    ytrain[:, idx],
                    early_stopping_rounds=early_stopping_rounds,
                    eval_set=[(xvalid, yvalid[:, idx])],
                    verbose=False,
                )
                trained_models.append(_m)
                if model_config.problem_type == ProblemType.multi_column_regression:
                    ypred_temp = _m.predict(xvalid)
                    if model_config.test_filename is not None:
                        test_pred_temp = _m.predict(xtest)
                else:
                    ypred_temp = _m.predict_proba(xvalid)[:, 1]
                    if model_config.test_filename is not None:
                        test_pred_temp = _m.predict_proba(xtest)[:, 1]

                ypred.append(ypred_temp)
                if model_config.test_filename is not None:
                    test_pred.append(test_pred_temp)

            ypred = np.column_stack(ypred)
            if model_config.test_filename is not None:
                test_pred = np.column_stack(test_pred)
            joblib.dump(
                trained_models,
                os.path.join(
                    model_config.output,
                    f"avf_model.{fold}",
                ),
            )

        else:
            model.fit(
                xtrain,
                ytrain,
                early_stopping_rounds=early_stopping_rounds,
                eval_set=[(xvalid, yvalid)],
                verbose=False,
            )

            joblib.dump(
                model,
                os.path.join(
                    model_config.output,
                    f"avf_model.{fold}",
                ),
            )

            if use_predict_proba:
                ypred = model.predict_proba(xvalid)
                if model_config.test_filename is not None:
                    test_pred = model.predict_proba(xtest)
            else:
                ypred = model.predict(xvalid)
                if model_config.test_filename is not None:
                    test_pred = model.predict(xtest)

        final_valid_predictions.update(dict(zip(valid_ids, ypred)))
        if model_config.test_filename is not None:
            final_test_predictions.append(test_pred)

        # calculate metric
        metric_dict = metrics.calculate(yvalid, ypred)
        if model_config.test_filename is not None:
            metric_test_dict = test_metrics.calculate(ytest, test_pred)
            test_scores.append(metric_test_dict)
        scores.append(metric_dict)
        logger.info(f"Fold {fold} done!")

    mean_metrics = dict_mean(scores)
    if model_config.test_filename is not None:
        test_mean_metrics = dict_mean(test_scores)
        logger.info(f"Test Metrics: {test_mean_metrics}")
        with open("test_metrics.json", "w") as fp:
            json.dump(test_mean_metrics, fp)

    logger.info(f"Metrics: {mean_metrics}")
    with open("metrics.json", "w") as fp:
        json.dump(mean_metrics, fp)
    save_valid_predictions(
        final_valid_predictions, model_config, target_encoder, "oof_predictions.csv"
    )

    logger.info(f"Best Params: {best_params}")
    with open("best_params.json", "w") as fp:
        json.dump(best_params, fp)

    if model_config.test_filename is not None:
        save_test_predictions(
            final_test_predictions,
            model_config,
            target_encoder,
            test_ids,
            "test_predictions.csv",
        )
    else:
        logger.info("No test data supplied. Only OOF predictions were generated")



from argparse import ArgumentParser

from ..autovf import AutoVF
from ..enums import TaskType
from . import BaseCommand


def train_autovf_command_factory(args):
    return TrainAutoVFCommand(
        args.train_filename,
        args.idx,
        args.targets,
        args.task,
        args.output,
        args.features,
        args.num_folds,
        args.use_gpu,
        args.seed,
        args.test_filename,
        args.time_limit,
        args.fast,
    )


class TrainAutoVFCommand(BaseCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        _parser = parser.add_parser("train", help="Train a new model using AutoVF")
        _parser.add_argument(
            "--train_filename",
            help="Path to training file",
            required=True,
            type=str,
        )
        _parser.add_argument(
            "--test_filename",
            help="Path to test file",
            required=False,
            type=str,
            default=None,
        )
        _parser.add_argument(
            "--output",
            help="Path to output directory",
            required=True,
            type=str,
        )
        _parser.add_argument(
            "--task",
            help="User defined task type",
            required=False,
            type=str,
            default=None,
            choices=TaskType.list_str(),
        )
        _parser.add_argument(
            "--idx",
            help="ID column",
            required=False,
            type=str,
            default="id",
        )
        _parser.add_argument(
            "--targets",
            help="Target column(s). If there are multiple targets, separate by ';'",
            required=False,
            type=str,
            default="target",
        )
        _parser.add_argument(
            "--num_folds",
            help="Number of folds to use",
            required=False,
            type=int,
            default=5,
        )
        _parser.add_argument(
            "--features",
            help="Features to use, separated by ';'",
            required=False,
            type=str,
            default=None,
        )
        _parser.add_argument(
            "--use_gpu",
            help="Whether to use GPU for training",
            action="store_true",
            required=False,
        )
        _parser.add_argument(
            "--fast",
            help="Whether to use fast mode for tuning params. Only one fold will be used if fast mode is set",
            action="store_true",
            required=False,
        )
        _parser.add_argument(
            "--seed",
            help="Random seed",
            required=False,
            type=int,
            default=42,
        )
        _parser.add_argument(
            "--time_limit",
            help="Time limit for optimization",
            required=False,
            type=int,
            default=None,
        )

        _parser.set_defaults(func=train_autovf_command_factory)

    def __init__(
        self,
        train_filename,
        idx,
        targets,
        task,
        output,
        features,
        num_folds,
        use_gpu,
        seed,
        test_filename,
        time_limit,
        fast,
    ):
        self.train_filename = train_filename
        self.idx = idx
        self.targets = targets.split(";")
        self.task = task
        self.output = output
        self.features = features.split(";") if features else None
        self.num_folds = num_folds
        self.use_gpu = use_gpu
        self.seed = seed
        self.test_filename = test_filename
        self.time_limit = time_limit
        self.fast = fast

    def execute(self):
        avf = AutoVF(
            train_filename=self.train_filename,
            idx=self.idx,
            targets=self.targets,
            task=self.task,
            output=self.output,
            features=self.features,
            num_folds=self.num_folds,
            use_gpu=self.use_gpu,
            seed=self.seed,
            test_filename=self.test_filename,
            time_limit=self.time_limit,
            fast=self.fast,
        )
        avf.train()



from argparse import ArgumentParser

from ..predict import AutoVFPredict
from . import BaseCommand


def predict_autovf_command_factory(args):
    return PredictAutoVFCommand(args.model_path, args.test_filename, args.out_filename)


class PredictAutoVFCommand(BaseCommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        _parser = parser.add_parser("predict", help="Make predictions on any CSV file")
        _parser.add_argument(
            "--model_path", help="Path to model", required=True, type=str
        )
        _parser.add_argument(
            "--test_filename",
            help="Path to test file",
            required=False,
            type=str,
            default=None,
        )
        _parser.add_argument(
            "--out_filename", help="Path to output file", required=True, type=str
        )
        _parser.set_defaults(func=predict_autovf_command_factory)

    def __init__(self, model_path, test_filename, out_filename):
        self.model_path = model_path
        self.test_filename = test_filename
        self.out_filename = out_filename

    def execute(self):
        avf = AutoVFPredict(model_path=self.model_path)
        avf.predict_file(
            test_filename=self.test_filename, out_filename=self.out_filename
        )



import logging

handler = logging.StreamHandler()
handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(message)s"))
logger = logging.getLogger("autovf")
logger.addHandler(handler)
logger.setLevel(logging.INFO)



from dataclasses import dataclass
from functools import partial

import numpy as np
from sklearn import metrics as skmetrics

from .enums import ProblemType
import copy


@dataclass
class Metrics:
    problem_type: ProblemType

    def __post_init__(self):
        if self.problem_type == ProblemType.binary_classification:
            self.valid_metrics = {
                "auc": skmetrics.roc_auc_score,
                "logloss": skmetrics.log_loss,
                "f1": skmetrics.f1_score,
                "accuracy": skmetrics.accuracy_score,
                "precision": skmetrics.precision_score,
                "recall": skmetrics.recall_score,
            }
        elif self.problem_type == ProblemType.multi_class_classification:
            self.valid_metrics = {
                "logloss": skmetrics.log_loss,
                "accuracy": skmetrics.accuracy_score,
                "mlogloss": skmetrics.log_loss,
            }
        elif self.problem_type in (
            ProblemType.single_column_regression,
            ProblemType.multi_column_regression,
        ):
            self.valid_metrics = {
                "r2": skmetrics.r2_score,
                "mse": skmetrics.mean_squared_error,
                "mae": skmetrics.mean_absolute_error,
                "rmse": partial(skmetrics.mean_squared_error, squared=False),
                "rmsle": partial(skmetrics.mean_squared_log_error, squared=False),
            }
        elif self.problem_type == ProblemType.multi_label_classification:
            self.valid_metrics = {
                "logloss": skmetrics.log_loss,
            }
        else:
            raise Exception("Invalid problem type")

    def calculate(self, y_true, y_pred):
        metrics = {}
        for metric_name, metric_func in self.valid_metrics.items():
            if self.problem_type == ProblemType.binary_classification:
                if metric_name == "auc":
                    metrics[metric_name] = metric_func(y_true, y_pred[:, 1])
                elif metric_name == "logloss":
                    metrics[metric_name] = metric_func(y_true, y_pred)
                else:
                    metrics[metric_name] = metric_func(y_true, y_pred[:, 1] >= 0.5)
            elif self.problem_type == ProblemType.multi_class_classification:
                if metric_name == "accuracy":
                    metrics[metric_name] = metric_func(
                        y_true, np.argmax(y_pred, axis=1)
                    )
                else:
                    metrics[metric_name] = metric_func(y_true, y_pred)
            else:
                if metric_name == "rmsle":
                    temp_pred = copy.deepcopy(y_pred)
                    temp_pred[temp_pred < 0] = 0
                    metrics[metric_name] = metric_func(y_true, temp_pred)
                else:
                    metrics[metric_name] = metric_func(y_true, y_pred)
        return metrics



def get_params(trial, model_config):
    params = {
        "learning_rate": trial.suggest_float("learning_rate", 1e-2, 0.25, log=True),
        "reg_lambda": trial.suggest_float("reg_lambda", 1e-8, 100.0, log=True),
        "reg_alpha": trial.suggest_float("reg_alpha", 1e-8, 100.0, log=True),
        "subsample": trial.suggest_float("subsample", 0.1, 1.0),
        "colsample_bytree": trial.suggest_float("colsample_bytree", 0.1, 1.0),
        "max_depth": trial.suggest_int("max_depth", 1, 9),
        "early_stopping_rounds": trial.suggest_int("early_stopping_rounds", 100, 500),
        "n_estimators": trial.suggest_categorical("n_estimators", [7000, 15000, 20000]),
    }
    if model_config.use_gpu:
        params["tree_method"] = "gpu_hist"
        params["gpu_id"] = 0
        params["predictor"] = "gpu_predictor"
    else:
        params["tree_method"] = trial.suggest_categorical(
            "tree_method", ["exact", "approx", "hist"]
        )
        params["booster"] = trial.suggest_categorical("booster", ["gbtree", "gblinear"])
        if params["booster"] == "gbtree":
            params["gamma"] = trial.suggest_float("gamma", 1e-8, 1.0, log=True)
            params["grow_policy"] = trial.suggest_categorical(
                "grow_policy", ["depthwise", "lossguide"]
            )

    return params



from setuptools import find_packages, setup


with open("README.md") as f:
    long_description = f.read()

INSTALL_REQUIRES = [
    "fastapi==0.70.0",
    "numpy==1.21.3",
    "optuna==2.10.0",
    "pyarrow==6.0.0",
    "pydantic==1.8.2",
    "joblib==1.1.1",
    "pandas==1.3.4",
    "scikit-learn==1.0.1",
    "uvicorn==0.15.0",
    "xgboost==1.5.0",
    "google-cloud-aiplatform==1.25.0",
]

if __name__ == "__main__":
    setup(
        name="autovf",
        description="autovf: tuning xgboost with optuna",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Ali Cabukel",
        author_email="alicabukel@proton.me",
        url="https://github.com/alicabukel/autovf",
        license="Apache 2.0",
        package_dir={"": "src"},
        packages=find_packages("src"),
        entry_points={"console_scripts": ["autovf=autovf.cli.autovf:main"]},
        include_package_data=True,
        install_requires=INSTALL_REQUIRES,
        platforms=["linux", "unix"],
        python_requires=">=3.6",
    )
